/*********************************************************************
//	Class ColorInfo
//
//	This class is statically accessible and can be used for all
//	color/smilie rendering operations, as long as a Canvas can
//	be supplied.
*********************************************************************/
class ColorInfo extends Actor
	abstract;

#exec TEXTURE IMPORT NAME=Smilie FILE=Textures\smile2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Frown FILE=Textures\sad2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=PlainFace FILE=Textures\plain2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=ShockedFace FILE=Textures\shocked2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Biggrin FILE=Textures\biggrin2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Tongueout FILE=Textures\tongue2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Unsureface FILE=Textures\unsure2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Winkface FILE=Textures\wink2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=AngrySmile FILE=Textures\angrysmile.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=AngrySad FILE=Textures\angrysad.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=LipsSealed FILE=Textures\lipssealed.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Crying FILE=Textures\crying.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Blush FILE=Textures\blush.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Sunglasses FILE=Textures\Sunglasses2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Finger FILE=Textures\Finger.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Angry FILE=Textures\Angry2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Kiss FILE=Textures\Kiss2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Heart FILE=Textures\HeartIcon.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=Confused FILE=Textures\Confused2.bmp GROUP="" FLAGS=2 MIPS=OFF
#exec TEXTURE IMPORT NAME=EDMUnderline FILE=Textures\EDMUnderline.bmp GROUP="" FLAGS=2 MIPS=OFF

// Constant colors
// ColorSet[0] = Red
// ColorSet[1] = Blue
// ColorSet[2] = Green
// ColorSet[3] = Yellow
// ColorSet[4] = Purple
// ColorSet[5] = Orange
// ColorSet[6] = Pink
// ColorSet[7] = Brown
// ColorSet[8] = White
// ColorSet[9] = Black
Var Const Color ColorSet[10], TrueBlack, ScoreboardGreen, DeathRed, NormalWhite, AdminMsgColor, SpecSayColor, IDColor, MuteColor;

// Dynamic colors
Var Color FlashColor, RandomColor;

// Floats used for calculating random and flash times
Var float FlashFloat, RandomFloat;

Var Const String AcceptChar, EndEffect; //Strings used that define all accepted letters and end all effects
Var Const Texture SmileTex[19];		// The smilie textures
Var Const float SmileScale, SmileWidth; //Constant sizes for drawing the smilies

struct FormatInfo
{
	var bool bBold;		// Is this text bold?
	var bool bUnderline;	// Is this text underlined?
	var bool bNoRender;	// Should we not render this text?
	var bool bShowSmiles;	// Are smiles allowed here?
	var bool bShowTags;	// Are we displaying effect tags?
	var int StartOffset;	// If the text wraps, where should we start again?
};

// Used internally to modify flash and random color values
static Function StaticTick(float Delta)
{
	default.FlashFloat -= 2;
	if(default.FlashFloat <= 20)
		default.FlashFloat = 255;

	default.FlashColor.R = default.FlashFloat * 1.34;
	default.FlashColor.G = default.FlashFloat * 1.04;
	default.FlashColor.B = default.FlashFloat * 0.94;

	default.RandomFloat -= Delta;
	if(default.RandomFloat <= 0)
	{
		default.RandomFloat = 0.3;
		default.RandomColor.R = Int(FRand() * 205) + 50;
		default.RandomColor.B = Int(FRand() * 205) + 50;
		default.RandomColor.G = Int(FRand() * 205) + 50;
	}
}

//////////////////////////
//// USABLE FUNCTIONS ////
//////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Remove color effects from a string								//
/// ColorGone - The string to remove tags from								//
/// Filter - Any tags that should be left in. All tags should be separated by a space			//
/// bShowNecessaryTags - If true, tags that come after a ^C^ tag will be left in the returned string	//
/////// Return: The given string with effect tags removed as specified by the given parameters		//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function String RemoveColoring(String ColorGone, Optional String Filter, Optional bool bShowNecessaryTags)
{
	local int CaretSpot;
	local bool bIncluding;
	local String TempStr, Colorless;

	Filter = Caps(Filter);

	while(ColorGone != "")
	{
		CaretSpot = InStr(ColorGone, "^");
		if(CaretSpot != -1)
		{
			Colorless = Colorless$Left(ColorGone, CaretSpot);

			TempStr = GrabColorSegment(Mid(ColorGone, CaretSpot));

			if(TempStr ~= "^/C^")
				bIncluding = false;

			if(!bIncluding && (Filter == "" || InStr(Filter, Caps(TempStr)) == -1))
			{
				if(ValidLet(TempStr) != "")
					CaretSpot += Len(TempStr);
				else if(ColorCoded(TempStr) != -1)
					CaretSpot += 3;
				else
				{
					Colorless = Colorless$"^";
					++CaretSpot;
				}
			}
			else
			{
				Colorless = Colorless$TempStr;
				CaretSpot += Len(TempStr);
			}

			if(TempStr ~= "^C^")
				bIncluding = bShowNecessaryTags;
		}
		else
		{
			Colorless = Colorless$ColorGone;
			CaretSpot = Len(ColorGone);
		}

		ColorGone = Mid(ColorGone, CaretSpot);
	}

	return Colorless;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Make sure the colors in someone's name do not affect the rest of a message (i.e. DeathMessages, name changes, enter/leave, etc) //
/// ColorGone - The text to correct													//
/////// Return: The given string with appropriate effect-ending tags placed at the end of a player's name				//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Insert tags to remove text effects after a player name is found									//
// --> Player name locations are determined by the first space in a string								//
//   after a text effect is noticed.													//
// --> Used for death messages, enter/leave messages, name change msgs, etc..								//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function string CorrectColors(String ColorGone)
{
	local int CaretSpot, SpcSpot;
	local String TempStr, ReturnString;

	ReturnString = "^f^";

	CaretSpot = InStr(ColorGone, "^");
	while(CaretSpot != -1)
	{
		TempStr = GrabColorSegment(Mid(ColorGone, CaretSpot));

		if(ColorCoded(TempStr) != -1 || ValidLet(TempStr) != "")
		{
			SpcSpot = InStr(ColorGone, " ");
			if(SpcSpot != -1)
			{
				ReturnString = ReturnString$Left(ColorGone, SpcSpot)@"^/C^^N^^F^";
				ColorGone = Mid(ColorGone, SpcSpot + 1);
			}
			else
				break;
		}
		else
		{
			ReturnString = ReturnString$Left(ColorGone, CaretSpot + 1);
			ColorGone = Mid(ColorGone, CaretSpot + 1);
		}

		CaretSpot = InStr(ColorGone, "^");
	}

	return ReturnString$ColorGone;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Render a piece of text that contains effect tags and smiles					//
/// Canvas - The Canvas											//
/// X - The X coordinate										//
/// Y - The Y coordinate										//
/// Message - The text to render									//
/// StartOffset - The position to start at again if the text wraps					//
/// bFade - Whether the text is fading									//
/// FadeCount - The degree to which the text is fading							//
/// bShowSmiles - Whether smiles are allowed here							//
/// Requestor - The UWindowWindow that is drawing this, if any						//
/// bNoRender - Whether to just calculate the lines and not render anything				//
/// bShowTags - Whether the effect tags should be left in tact						//
/////// Return: The number of lines that are taken up to render this text with the given parameters	//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function int RenderColoredString(Canvas Canvas, float X, float Y, String Message, optional int StartOffset, optional bool bFade, optional float FadeCount, optional bool bShowSmiles, optional UWindowWindow Requestor, optional bool bNoRender, optional bool bShowTags)
{
	local int CaretSpot, LineCount, Result;
	local String TempStr, ChunkStr;
	local Font OrigFont;
	local Color OrigColor, CurColor;
	local Byte OrigStyle;
	local FormatInfo Formatting;

	OrigFont = Canvas.Font;
	OrigStyle = Canvas.Style;
	OrigColor = Canvas.DrawColor;
	Formatting.StartOffset = StartOffset;
	Formatting.bNoRender = bNoRender;
	Formatting.bShowTags = bShowTags;
	Formatting.bShowSmiles = (bShowSmiles && class'EDMSettingsKeeper'.default.bRenderFaces);

	if(bFade)
		Canvas.DrawColor = FadeColor(OrigColor, FadeCount);
	else
		Canvas.Style = ERenderStyle.STY_Normal;

	CurColor = Canvas.DrawColor;
	LineCount = 1;

	CaretSpot = InStr(Message, "^");
	while(CaretSpot != -1)
	{
		TempStr = GrabColorSegment(Mid(Message, CaretSpot));

		Result = ColorCoded(TempStr);
		if(Result == -1)
		{
			TempStr = ValidLet(TempStr);
			if(TempStr == "")
				++CaretSpot;
			else
			{
				if(!Formatting.bShowTags) // If we're not displaying the tags
				{
					if(TempStr ~= "^R^")
					{
						if(bFade)
							CurColor = FadeColor(default.RandomColor, FadeCount);
						else
							CurColor = default.RandomColor;
					}
					else if(TempStr ~= "^L^") 
					{
						CurColor.R = Min(255, CurColor.R + 15);
						CurColor.B = Min(255, CurColor.B + 15);
						CurColor.G = Min(255, CurColor.G + 15);

						if(bFade) CurColor = FadeColor(CurColor, FadeCount);
					}
					else if(TempStr ~= "^D^") 
					{
						CurColor.R = Max(0, CurColor.R - 15);
						CurColor.B = Max(0, CurColor.B - 15);
						CurColor.G = Max(0, CurColor.G - 15);

						if(bFade) CurColor = FadeColor(CurColor, FadeCount);
					}
					else if(TempStr ~= "^I^") 
					{
						CurColor.R = CurColor.R ^ 255;
						CurColor.B = CurColor.B ^ 255;
						CurColor.G = CurColor.G ^ 255;

						if(bFade) CurColor = FadeColor(CurColor, FadeCount);
					}
					else if(TempStr ~= "^FLASH^")
					{
						if(bFade)
							CurColor = FadeColor(default.FlashColor, FadeCount);
						else
							CurColor = default.FlashColor;
					}

					// Don't try to be a smartass and put the fade here. If you do that, the text may get faded
					// multiple times.
				}
				else
				{
					// If this function was called with bShowTags as true, then that supercedes the formatting
					// setting. All tags should always be shown if bShowTags is true
					if(bShowTags || (!(TempStr ~= "^/C^" || TempStr ~= "^F^" || TempStr ~= "^/F^")))
					{
						TempStr = "";
						++CaretSpot;
					}
				}
			}
		}
		else
		{
			if(!Formatting.bShowTags) // If we're not displaying the tags
			{
				if(bFade)
					CurColor = FadeColor(default.ColorSet[Result], FadeCount);
				else
					CurColor = default.ColorSet[Result];
			}
			else
			{
				TempStr = "";
				++CaretSpot;
			}
		}

		ChunkStr = Left(Message, CaretSpot);

		if(Formatting.bShowSmiles)
			LineCount += HandleSmileChunk(Canvas, ChunkStr, Formatting, X, Y, bFade, FadeCount, Requestor);
		else
			LineCount += RenderColoredText(Canvas, ChunkStr, Formatting, X, Y, Requestor);

		if(class'EDMSettingsKeeper'.default.bRenderColors && !(TempStr ~= "^N^"))
		{
			if(TempStr ~= "^S^" && Requestor == None)
				Canvas.Font = Canvas.SmallFont;
			else if(TempStr ~= "^/S^")
				Canvas.Font = OrigFont;
			else if(TempStr ~= "^M^" && Requestor == None)
				Canvas.Font = Font'MedFont';
			else if(TempStr ~= "^/M^")
				Canvas.Font = OrigFont;
			else if(TempStr ~= "^B^")
				Formatting.bBold = true;
			else if(TempStr ~= "^/B^")
				Formatting.bBold = false;
			else if(TempStr ~= "^U^")
				Formatting.bUnderline = true;
			else if(TempStr ~= "^/U^")
				Formatting.bUnderline = false;
			else if(TempStr ~= "^T^")
				Canvas.Style = ERenderStyle.STY_Translucent;
			else if(TempStr ~= "^/T^")
				Canvas.Style = OrigStyle;
		}
		else
		{
			if(bFade)
				CurColor = FadeColor(OrigColor, FadeCount);
			else
				CurColor = OrigColor;
			Formatting.bBold = false;
			Formatting.bUnderline = false;
			Canvas.Style = OrigStyle;
			Canvas.Font = OrigFont;
		}

		if(TempStr ~= "^C^")
			Formatting.bShowTags = true;
		else if(TempStr ~= "^/C^")
			Formatting.bShowTags = bShowTags; // Don't hide tags if bShowTags is true
		else if(TempStr ~= "^F^")
			Formatting.bShowSmiles = false;
		else if(TempStr ~= "^/F^")
			Formatting.bShowSmiles = (bShowSmiles && class'EDMSettingsKeeper'.default.bRenderFaces); // Don't show smiles if bShowSmiles is false or if the client doesn't want them

		Message = Mid(Message, CaretSpot + Len(TempStr));

		Canvas.DrawColor = CurColor;

		CaretSpot = InStr(Message, "^");
	}

	if(Formatting.bShowSmiles)
		LineCount += HandleSmileChunk(Canvas, Message, Formatting, X, Y, bFade, FadeCount, Requestor);
	else
		LineCount += RenderColoredText(Canvas, Message, Formatting, X, Y, Requestor);

	Canvas.DrawColor = OrigColor;
	Canvas.Style = OrigStyle;
	Canvas.Font = OrigFont;

	return LineCount;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Like the Left() String function, except you can specify how far to go when taking tags into consideration  //
/// ColorGone - The string to get the left() of									   //
/// HowFar - How far to go when effect tags are removed								   //
/// HowFarTotal - How far to go when effect tags are counted							   //
/////// Return: The Left() of a string with the parameters taken into consideration				   //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function String ColorLeft(String ColorGone, int HowFar, int HowFarTotal)
{
	local int i;
	local String TempStr, ResultStr;

	while( Len(ResultStr) < HowFarTotal && Len(RemoveColoring(ResultStr, "", true)) < HowFar
		&& i < Len(ColorGone) )
	{
		if(Len(ColorGone) - i >= 4)
			TempStr = GrabColorSegment(Mid(ColorGone, i));
		else
			TempStr = Mid(ColorGone, i, 3);

		if(ColorCoded(TempStr) != -1 || ValidLet(TempStr) != "")
		{
			ResultStr = ResultStr$TempStr;
			i += Len(TempStr);
		}
		else
			ResultStr = ResultStr$Mid(ColorGone, i++, 1);
	}

	return ResultStr;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Find the true XL of a string with font modifiers.							//
/// Canvas - The canvas												//
/// StrCheck - The string to check. This string should have had its color's removed with a "^S^ ^M^ ^N^" filter //
/////// Return: The horizontal space this string would take up if rendered with the font modifiers              //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function float TrueStringSize(Canvas Canvas, String StrCheck)
{
	local String temp;
	local int index;
	local Font f;
	local float XL, YL, TotalSize;

	f = Canvas.Font;
	if(f == None || StrCheck == "")
		return 0;

	if(!class'EDMSettingsKeeper'.default.bRenderColors)
	{
		Canvas.StrLen(RemoveColoring(StrCheck, "", true), XL, YL);
		return XL;
	}
	else
	{
		index = FindNextFontIndex(StrCheck);
		while(index != -1)
		{
			temp = Left(StrCheck, index);
			Canvas.StrLen(temp, XL, YL);
			TotalSize += XL;
			temp = Mid(StrCheck, index, 3);
			if(temp ~= "^S^")
				Canvas.Font = Font'SmallFont';
			else if(temp ~= "^M^")
				Canvas.Font = Font'MedFont';
			else if(temp ~= "^N^")
				Canvas.Font = f;

			StrCheck = Mid(StrCheck, index + 3);

			index = FindNextFontIndex(StrCheck);
		}

		Canvas.StrLen(StrCheck, XL, YL);
		TotalSize += XL;

		Canvas.Font = f;

		return TotalSize;
	}
}



////////////////////////////
//// HELPER FUNCTIONS //////
////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Test this string for an effect tag, and return whatever is found.				//
/// ColorGrab - The string that will be checked for an effect tag					//
/////// Return: The portion of the given string that most likely contains an effect. There is no	//
///////		certainty at this point that it actually does, though. 					//
//////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function String GrabColorSegment(String ColorGrab)
{
	if(Left(ColorGrab, 7) ~= "^FLASH^")
		return Left(ColorGrab, 7);
	else
	{
		ColorGrab = Left(ColorGrab, 4);
		if(Right(ColorGrab, 1) != "^" || Mid(ColorGrab, 2, 1) == "^")
			ColorGrab = Left(ColorGrab, 3);
		return ColorGrab;
	}
}

////////////////////////////////////////////////////////////////////////////
/////// Render a smile texture						  //
/// Canvas - The Canvas							  //
/// SmileTex - The texture of the smile					  //
/// Format - The current rendering format				  //
/// X - The X coordinate						  //
/// Y - The Y coordinate						  //
/// bFade - Whether the text is fading					  //
/// FadeCount - The degree to which the text is fading			  //
/// Requestor - The UWindowWindow that is drawing this, if any		  //
/////// Return: The number of times the text wrapped to render this smile //
////////////////////////////////////////////////////////////////////////////
static function int RenderSmile(Canvas Canvas, Texture SmileTex, FormatInfo Format, out float X, out float Y, Optional bool bFade, Optional float FadeCount, Optional UWindowWindow Requestor)
{
	local float XL, YL;
	local int LineCount;
	local Color TempColor;

	if(X + default.SmileWidth > Canvas.ClipX) // In case the smile will go off the screen
	{
		if(Requestor == None)
		{
			X = Format.StartOffset;
			Canvas.StrLen("TEST", XL, YL);
		}
		else
		{
			X = 2;
			Requestor.TextSize(Canvas, "TEST", XL, YL);
		}
		Y += YL;
		++LineCount;
	}

	TempColor = Canvas.DrawColor;

	if(bFade)
		Canvas.DrawColor = FadeColor(default.ColorSet[8], FadeCount);
	else
		Canvas.DrawColor = default.ColorSet[8];

	if(!Format.bNoRender)
	{
		if(Requestor == None)
		{
			Canvas.SetPos(X, Y);
			Canvas.DrawIcon(SmileTex, default.SmileScale);
		}
		else
			Requestor.DrawStretchedTexture(Canvas, X, Y, SmileTex.USize*default.SmileScale, SmileTex.VSize*default.SmileScale, SmileTex);
	}

	Canvas.DrawColor = TempColor;
	X += default.SmileWidth;

	return LineCount;
}

////////////////////////////////////////////////////////////////////////////////////
/////// Render some text while taking wrapping and formatting into consideration  //
/// Canvas - The Canvas								  //
/// Message - The text to render						  //
/// Format - The text formatting						  //
/// XSpot - The X coordinate							  //
/// Height - The Y coordinate							  //
/// Requestor - The UWindowWindow that is drawing this, if any			  //
/////// Return: The number of times the text was wrapped to render the given text //
////////////////////////////////////////////////////////////////////////////////////
static function int RenderColoredText(Canvas Canvas, string Message, FormatInfo Format, out float XSpot, out float Height, Optional UWindowWindow Requestor)
{
	local float XL, YL, TestYL, MaxWidth;
	local int i, LineCount, StartOffset, spcSpot;
	local string TempString;
	local bool bResized;
	local Font f;

	if(Requestor == None)
	{
		f = Canvas.Font;
		StartOffset = Format.StartOffset;
		Canvas.Font = Font'WhiteFont';
		Canvas.StrLen("TEST", XL, TestYL);
		Canvas.Font = f;
	}
	else
	{
		Requestor.TextSize(Canvas, "TEST", XL, TestYL);
		StartOffset = 2;
	}
	
	while(Message != "")
	{
		spcSpot = InStr(Message, " ");

		if(spcSpot != -1)
			TempString = Left(Message, spcSpot + 1);
		else
			TempString = Message;

		if(Requestor == None)
			Canvas.StrLen(TempString, XL, YL);
		else
			Requestor.TextSize(Canvas, TempString, XL, YL);

		// If this text will not fit on the screen from our current position
		if(XSpot + XL > Canvas.ClipX)
		{
			// If this text can simply be pushed onto the next line, then do it
			if(StartOffset + XL < Canvas.ClipX)
			{
				Height += TestYL;
				XSpot = StartOffset;
				++LineCount;
			}
			else
			{
				// This is an unfortunately slow and expensive process, but it's the only conceivable way to
				// find the first position within the string that will allow it to fit on the screen. When we
				// find the spot in TempString that will allow it to fit on the screen, we insert a space so
				// that the next time the loop iterates, the space will be detected and the string will be broken
				// from there.
				bResized = false;
				i = Len(TempString);
				while(i > 0 && !bResized)
				{
					if(Requestor == None)
						Canvas.StrLen(Left(TempString, i)$" ", XL, YL);
					else
						Requestor.TextSize(Canvas, Left(TempString, i)$" ", XL, YL);

					if(StartOffset + XL < Canvas.ClipX)
					{
						Message = Left(TempString, i)@Mid(TempString, i)$Mid(Message, Len(TempString));
						bResized = true;
					}
					else
						--i;
				}

				// If we couldn't resize it, forget it. The canvas isn't even big enough to display one character
				if(!bResized)
					Message = "";
			}
		}
		else
		{
			Message = Mid(Message, Len(TempString));
			RenderIt(Canvas, TempString, Format, XSpot, Height, Requestor);
			XSpot += XL;
		}
	}

	return LineCount;
}

/////////////////////////////////////////////////////////////////////////////////////
/////// Render a piece of text in a given position with bold/underline formatting  //
/// Canvas - The canvas                                                            //
/// Message - The text to render                                                   //
/// Format - The text formatting                                                   //
/// XSpot - The X coordinate                                                       //
/// Height - The Y coordinate                                                      //
/// Requestor - The UWindowWindow that is drawing this, if any                     //
/////////////////////////////////////////////////////////////////////////////////////
static Function RenderIt(Canvas Canvas, String Message, FormatInfo Format, float XSpot, float Height, Optional UWindowWindow Requestor)
{
	local float XL, YL;

	if(!Format.bNoRender)
	{
		if(Requestor == None)
		{
			Canvas.StrLen(Message, XL, YL);
			Canvas.SetPos(XSpot, Height);
		}
		else
			Requestor.TextSize(Canvas, Message, XL, YL);
	
		if(Format.bBold)
			RenderBold(Canvas, Message, XSpot, Height, Requestor);
		else
		{
			if(Requestor == None)
				Canvas.DrawText(Message);
			else
				Requestor.ClipText(Canvas, XSpot, Height, Message);
		}
	
		if(Format.bUnderline)
		{
			if(Requestor == None)
				Height += (YL * 0.9);
			else
				Height += (YL * 0.8);
	
			Canvas.SetPos(XSpot, Height);
	
			if(Requestor == None)
				Canvas.DrawTile(Texture'EDMUnderline',XL,2.0,0,0,32.0,2.0); //Draw line under the text
			else
				Requestor.DrawStretchedTexture(Canvas, XSpot, Height, XL, 2.0, Texture'EDMUnderLine');
		}
	}
}


/////////////////////////////////////////////////////////////
////// Render a piece of text as bold text 		   //
/// Canvas - The Canvas					   //
/// S - The text to draw				   //
/// X - The X coordinate				   //
/// Y - The Y coordinate                                   //
/// Requestor - The UWindowWindow drawing this, if any     //
/////////////////////////////////////////////////////////////
static Function RenderBold(Canvas Canvas, String S, float X, float Y, Optional UWindowWindow Requestor)
{
	if(Requestor == None)
	{
		Canvas.DrawText(S);
		Canvas.SetPos(X, Y + 0.01);
		Canvas.DrawText(S);
		Canvas.SetPos(X - 0.01, Y);
		Canvas.DrawText(S);
		Canvas.SetPos(X + 0.01, Y);
		Canvas.DrawText(S);
		Canvas.SetPos(X, Y - 0.01);
		Canvas.DrawText(S);
	}
	else
	{
		Requestor.ClipText(Canvas, X, Y + 0.01, S);
		Requestor.ClipText(Canvas, X - 0.01, Y, S);
		Requestor.ClipText(Canvas, X + 0.01, Y, S);
		Requestor.ClipText(Canvas, X, Y - 0.01, S);
	}
}

///////////////////////////////////////////////////////////////
////// Fade a color to a given degree			     //
/// TempColor - The original color			     //
/// FadeFactor - The degree to fade it (between 0.0 and 1.0) //
////// Return: The faded color				     //
///////////////////////////////////////////////////////////////
static Function Color FadeColor(Color TempColor, float FadeFactor)
{
	TempColor.R = TempColor.R * FadeFactor;
	TempColor.G = TempColor.G * FadeFactor;
	TempColor.B = TempColor.B * FadeFactor;
	return TempColor;
}

/////////////////////////////////////////////////////////////////////////////////////////
////// Check a string to see if any known smiles are in it 			       //
/// SmileCheck - The string to search for smiles	   			       //
////// Return: The index of the smile that is in this string. -1 if no smile is found. //
/////////////////////////////////////////////////////////////////////////////////////////
Static Function int CheckForSmile(String SmileCheck)
{
	local String First, Last, Middle;

	SmileCheck = Caps(SmileCheck);

	First = Caps(Left(SmileCheck, 1));
	Last = Caps(Right(SmileCheck, 1));
	Middle = Caps(Mid(SmileCheck, 1, 1));

	// Check to see if this is an invalid smile
	if(Len(SmileCheck) > 2)
		if(First != ">" && Middle != "-" && Middle != "'" && Middle != "_")
			return -1;
		else if( (First == ">" || First == "D") && Middle != ":" && Middle != "=")
			return -1;

	if(First == ">")
	{
		if(Last == ")")
			return 0;
		else if(Last == "(")
			return 1;
	}
	else if(First == "D")
	{
		if(Last == ":" && Middle == ":")
			return 15;
	}
	else if((First == ";") && (Last == ")"))
		return 3;
	else if((First == "8") && (Last == ")"))
		return 13;
	else if((SmileCheck == ":'(") || (SmileCheck == "='(") || (SmileCheck == ";_;"))
		return 5;
	else if((First == ":") || (First == "="))
	{
		if(Last == ")")
			return 2;
		else if(Last == "(")
			return 4;
		else if(Last == "/" || Last == "\\")
			return 6;
		else if(Last == "D")
			return 7;
		else if(Last == "P")
			return 8;
		else if(Last == "O" || Last == "0")
			return 9;
		else if(Last == "I")
			return 10;
		else if(Last == "|")
			return 10;
		else if(Last == "X")
			return 11;
		else if(Last == "[")
			return 12;
		else if(Last == "F")
			return 14;
		else if(Last == "*")
			return 16;
		else if(Last == "H")
			return 17;
		else if(Last == "S")
			return 18;
	}

	return -1;
}

////////////////////////////////////////////////////////////////////////////////////////////
////// Check if a string is a valid effect						  //
/// ValCheck - The string to check							  //
////// Return: The effect text, if the string is valid. Blank if the string is not valid. //
////////////////////////////////////////////////////////////////////////////////////////////
static Function String ValidLet(String ValCheck)
{
	if(ValCheck ~= "^FLASH^")
		return ValCheck;
	else
	{
		if(Len(ValCheck) <= 4 && Len(ValCheck) >= 3)
		{
			if(Left(ValCheck, 1) == "^" && Right(ValCheck, 1) == "^")
			{
				ValCheck = Mid(ValCheck, 1, Len(ValCheck) - 2);
	
				if(Len(ValCheck) == 1 || Left(ValCheck, 1) == "/")
				{
					if(InStr(default.AcceptChar, Caps(ValCheck)) != -1)
						return "^"$ValCheck$"^";
					else
						return "";
				}
				else
					return "";
			}
			else
				return "";
		}
		else
			return "";
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////// Search one string for a set of delimiters and return the index of the first one 		   //
/// src - The string to search										   //
/// delimiters - The set of delimiters to look for							   //
/////// Return: The index of the earliest detected delimiter. Returns the length of src if none are found. //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
static Function int StrCSpn(String src, String delimiters)
{
	local int i;
	local int index, temp;
	index = Len(src);
	for(i = 0; i < Len(delimiters); ++i)
	{
		temp = InStr(src, Mid(delimiters, i, 1));
		if(temp < index && temp != -1)
			index = temp; 
	}
	return index;
}

//////////////////////////////////////////////////////////////////////////////////////////
/////// Render a piece of text that may contain smiles					//
/// Canvas - The Canvas									//
/// TempString - The String to render							//
/// Format - The current text formatting						//
/// XSpot - The X coordinate								//
/// Height - The Y coordinate								//
/// bFade - Whether the text is being faded						//
/// FadeCount - The degree that the text is faded					//
/// Requestor - The UWindowWindow that is rendering this, if any			//
/////// Returns: The number of times the text was wrapped to render the given string	//
//////////////////////////////////////////////////////////////////////////////////////////
static Function int HandleSmileChunk(Canvas Canvas, String TempString, FormatInfo Format, Out float XSpot, Out float Height, Optional bool bFade, Optional float FadeCount, Optional UWindowWindow Requestor)
{
	local int index, SmileNum, SmileLen, LineCount;

	while(TempString != "")
	{
		index = StrCSpn(TempString, ":=;8");
		SmileLen = 0;
		if(index < Len(TempString))
		{
			if(index > 0)
				if(Mid(TempString, index - 1, 1) == ">" || Mid(TempString, index - 1, 1) == "D")
				{
					--index;
					SmileLen = 1;
				}

			if(Len(Mid(TempString, index)) >= 2)
			{
				SmileNum = CheckForSmile(Mid(TempString, index, 2));
				if(SmileNum == -1)
				{
					if(Len(Mid(TempString, index)) >= 3)
					{
						SmileNum = CheckForSmile(Mid(TempString, index, 3));
						if(SmileNum == -1)
							++SmileLen;
						else
							SmileLen = 3;
					}
					else
					{
						SmileNum = -1;
						++SmileLen;
					}
				}
				else
					SmileLen = 2;
			}
			else
			{
				SmileNum = -1;
				++SmileLen;
			}
		}
		else
		{
			SmileLen = 0;
			SmileNum = -1;
		}

		if(SmileNum != -1)
		{
			LineCount += RenderColoredText(Canvas, Left(TempString, index), Format, XSpot, Height, Requestor);
			LineCount += RenderSmile(Canvas, default.SmileTex[SmileNum], Format, XSpot, Height, bFade, FadeCount, Requestor);
		}
		else
			LineCount += RenderColoredText(Canvas, Left(TempString, index + SmileLen), Format, XSpot, Height, Requestor);

		TempString = Mid(TempString, index + SmileLen);
	}

	return LineCount;
}

///////////////////////////////////////////////////////////////////////////////////////////
/////// Check if a string has a any numeric color codes in it				 //
/// ColCheck - The string to check							 //
////// Return: The number of the color code that is found. -1 if no color code is found. //
///////////////////////////////////////////////////////////////////////////////////////////
static Function int ColorCoded(String ColCheck)
{
	if(Len(ColCheck) == 3)
	{
		if(Left(ColCheck, 1) == "^" && Right(ColCheck, 1) == "^")
		{
			if(InStr("0123456789", Mid(ColCheck, 1, 1)) != -1)
				return int(Mid(ColCheck, 1, 1));
			else
				return -1;
		}
		else
			return -1;
	}
	else
		return -1;
}

/////////////////////////////////////////////////////////////////////////////////////
/////// Find the first occurrence of a font modifying tag in a given string	   //
/// StrCheck - The string to check for font modifying tags			   //
/////// Return: The index of the first font modifying tag, or -1 if none was found //
/////////////////////////////////////////////////////////////////////////////////////
static Function int FindNextFontIndex(String StrCheck)
{
	local int index, tmpIndex;

	index = InStr(Caps(StrCheck), "^S^");
	tmpIndex = InStr(Caps(StrCheck), "^M^");
	if( index == -1 || (tmpIndex < index && tmpIndex != -1) )
		index = tmpIndex;
	tmpIndex = InStr(Caps(StrCheck), "^N^");
	if( index == -1 || (tmpIndex < index && tmpIndex != -1) )
		index = tmpIndex;

	return index;
}

defaultproperties
{
				ColorSet(0)=(R=255,G=30,B=30)
				ColorSet(1)=(R=10,G=10,B=255)
				ColorSet(2)=(R=30,G=255,B=30)
				ColorSet(3)=(R=200,G=200,B=30)
				ColorSet(4)=(R=200,G=30,B=200)
				ColorSet(5)=(R=255,G=130,B=15)
				ColorSet(6)=(R=250,G=125,B=125)
				ColorSet(7)=(R=145,G=99,B=55)
				ColorSet(8)=(R=255,G=255,B=255)
				ColorSet(9)=(R=20,G=20,B=20)
				ScoreboardGreen=(G=255)
				DeathRed=(R=255,G=10,B=10)
				NormalWhite=(R=210,G=210,B=210)
				AdminMsgColor=(R=100,G=190,B=120)
				SpecSayColor=(R=140,G=213,B=56)
				IDColor=(R=135,G=160,B=70)
				MuteColor=(R=90,G=190,B=70)
				AcceptChar="/C/M/U/B/F/T/SIDNRL"
				EndEffect="^/C^^N^^/F^"
				SmileTex(0)=Texture'JRB.AngrySmile'
				SmileTex(1)=Texture'JRB.AngrySad'
				SmileTex(2)=Texture'JRB.Smilie'
				SmileTex(3)=Texture'JRB.Winkface'
				SmileTex(4)=Texture'JRB.Frown'
				SmileTex(5)=Texture'JRB.Crying'
				SmileTex(6)=Texture'JRB.Unsureface'
				SmileTex(7)=Texture'JRB.Biggrin'
				SmileTex(8)=Texture'JRB.Tongueout'
				SmileTex(9)=Texture'JRB.ShockedFace'
				SmileTex(10)=Texture'JRB.PlainFace'
				SmileTex(11)=Texture'JRB.LipsSealed'
				SmileTex(12)=Texture'JRB.Blush'
				SmileTex(13)=Texture'JRB.Sunglasses'
				SmileTex(14)=Texture'JRB.Finger'
				SmileTex(15)=Texture'JRB.Angry'
				SmileTex(16)=Texture'JRB.Kiss'
				SmileTex(17)=Texture'JRB.Heart'
				SmileTex(18)=Texture'JRB.Confused'
				SmileScale=0.350750
				SmileWidth=15.200000
}
